/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_REDUCEMEANIMPL_FORWARD_KERNEL_H_
#define AIDGE_CPU_OPERATOR_REDUCEMEANIMPL_FORWARD_KERNEL_H_

#include "aidge/utils/Registrar.hpp"
#include "aidge/operator/ReduceMean.hpp"
#include "aidge/backend/cpu/operator/ReduceMeanImpl.hpp"
#include <array>
#include <cstddef>
#include <algorithm>
#include "aidge/data/Data.hpp"

namespace Aidge {
template <class I, class O, DimSize_t DIM>
void ReduceMeanImpl_cpu_forward_kernel(const typename ReduceMean_Op<DIM>::Attrs& attrs,
                                     const std::vector<DimSize_t>& inputDims,
                                     const void* input_,
                                     void* output_) {

    const I* input = static_cast<const I*>(input_);
    O* output = static_cast<O*>(output_);

	DimSize_t keepDims = std::get<1>(attrs);
    // Calculate the total number of elements in the input array
    size_t totalElements = 1;
    for (size_t dimSize : inputDims) {
        totalElements *= dimSize;
    }

    // Create a temporary arrays to store intermediate input/output for each Reduce op
    std::vector<I> tempInArray(input, input + totalElements);
    std::vector<I> tempOutArray(input, input + totalElements);
    std::vector<size_t> currentDims = inputDims;

    std::size_t addedElems = 0;
    for(std::size_t i=0; i<DIM ; ++i)
    {
		addedElems = 0;
		int axis_ = std::get<0>(attrs)[i];
		std::size_t axis = axis_>=0? axis_: axis_ + inputDims.size();

		I* tempOutArrayPtr = tempOutArray.data();

		std::size_t postAxisElems = 1;
		for (std::size_t d = axis + 1; d < inputDims.size(); ++d) {
			postAxisElems *= inputDims[d];
		}
		std::size_t preAxisElems = 1;
		for (std::size_t d = 0; d < axis; ++d) {
			preAxisElems *= inputDims[d];
		}

        for (std::size_t j=0; j<preAxisElems; ++j)
        {
            for (std::size_t k=0; k<postAxisElems; ++k)
            {
				// Compute the mean value for the element k of each stride
                I mean = 0;
                for(std::size_t l=0; l<currentDims[axis];l++)
                {
					size_t idx = j * (postAxisElems * currentDims[axis]) + l * postAxisElems + k;
					mean += tempInArray[idx];
                }
                tempOutArrayPtr[addedElems] = mean / currentDims[axis];
                addedElems++;
            }
        }

        // Update the input for the next reduce operation
        tempInArray.assign(tempOutArray.begin(), tempOutArray.begin() + addedElems);
        if(keepDims)
			currentDims[axis] = 1;
        else if (currentDims.size()>1)
			currentDims.erase(currentDims.begin()+axis);
    }
	std::copy_n(tempInArray.data(), addedElems, output);
}
namespace {
// DIM = 1
static Registrar<ReduceMeanImpl1DForward_cpu> registrarReduceMeanImplForward_1D_cpu_Float32(
        {DataType::Float32, DataType::Float32}, Aidge::ReduceMeanImpl_cpu_forward_kernel<float, float,1>);
static Registrar<ReduceMeanImpl1DForward_cpu> registrarReduceMeanImplForward_1D_cpu_Int32(
        {DataType::Int32, DataType::Int32}, Aidge::ReduceMeanImpl_cpu_forward_kernel<int, int,1>);
static Registrar<ReduceMeanImpl1DForward_cpu> registrarReduceMeanImplForward_1D_cpu_Float64(
        {DataType::Float64, DataType::Float64}, Aidge::ReduceMeanImpl_cpu_forward_kernel<double, double,1>);

// DIM = 2
static Registrar<ReduceMeanImpl2DForward_cpu> registrarReduceMeanImplForward_2D_cpu_Float32(
        {DataType::Float32, DataType::Float32}, Aidge::ReduceMeanImpl_cpu_forward_kernel<float, float,2>);
static Registrar<ReduceMeanImpl2DForward_cpu> registrarReduceMeanImplForward_2D_cpu_Int32(
        {DataType::Int32, DataType::Int32}, Aidge::ReduceMeanImpl_cpu_forward_kernel<int, int,2>);
static Registrar<ReduceMeanImpl2DForward_cpu> registrarReduceMeanImplForward_2D_cpu_Float64(
        {DataType::Float64, DataType::Float64}, Aidge::ReduceMeanImpl_cpu_forward_kernel<double, double,2>);

// DIM = 3
static Registrar<ReduceMeanImpl3DForward_cpu> registrarReduceMeanImplForward_3D_cpu_Float32(
        {DataType::Float32, DataType::Float32}, Aidge::ReduceMeanImpl_cpu_forward_kernel<float, float,3>);
static Registrar<ReduceMeanImpl3DForward_cpu> registrarReduceMeanImplForward_3D_cpu_Int32(
        {DataType::Int32, DataType::Int32}, Aidge::ReduceMeanImpl_cpu_forward_kernel<int, int,3>);
static Registrar<ReduceMeanImpl3DForward_cpu> registrarReduceMeanImplForward_3D_cpu_Float64(
        {DataType::Float64, DataType::Float64}, Aidge::ReduceMeanImpl_cpu_forward_kernel<double, double,3>);
}  // namespace
}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_REDUCEMEANIMPL_FORWARD_KERNEL_H_ */
